import React, { useState, useEffect, useCallback, useMemo, Fragment, useRef, useImperativeHandle } from 'react';
import styled, { keyframes, css, ThemeProvider } from 'styled-components';
import { areEqual, FixedSizeList } from 'react-window';

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

const fadeIn = keyframes`
  from {
    opacity: 0;
  } to {
    opacity: 1;
  }
`;
const fadeInAnimationCss = css`
  animation: ${fadeIn} 0.225s ease-in-out forwards;
`;
const mediaQueryIsIE = '@media all and (-ms-high-contrast: none), (-ms-high-contrast: active)';

const useDebounce = (value, delay) => {
  const [debouncedValue, setDebouncedValue] = useState(value);
  useEffect(() => {
    if (delay === undefined) {
      return undefined;
    }

    const handler = setTimeout(() => {
      setDebouncedValue(value);
    }, delay);
    return () => {
      clearTimeout(handler);
    };
  }, [value, delay]);
  return delay === undefined ? value : debouncedValue;
};

function isDocumentElement(el) {
  return el === document.documentElement || el === document.body || el === window;
}

function easeOutCubic(t, s, c, d) {
  return c * ((t = t / d - 1) * t * t + 1) + s;
}

function getScrollTop(el) {
  return isDocumentElement(el) ? window.pageYOffset : el.scrollTop;
}

function scrollTo(el, top) {
  if (isDocumentElement(el)) {
    window.scrollTo(0, top);
  } else {
    el.scrollTop = top;
  }
}

function getScrollParent(el) {
  let style = getComputedStyle(el);
  const excludeStaticParent = style.position === 'absolute';

  if (style.position === 'fixed') {
    return document.documentElement;
  }

  for (let parent = el; parent = parent.parentElement;) {
    style = getComputedStyle(parent);

    if (excludeStaticParent && style.position === 'static') {
      continue;
    } else if (/(auto|scroll)/.test(`${style.overflow}${style.overflowY}${style.overflowX}`)) {
      return parent;
    }
  }

  return document.documentElement;
}

function smoothScrollTo(element, to, duration, callback) {
  const start = getScrollTop(element);
  const change = to - start;
  let currentTime = 0;

  function smoothScroller() {
    currentTime += 5;
    scrollTo(element, easeOutCubic(currentTime, start, change, duration));

    if (currentTime < duration) {
      window.requestAnimationFrame(smoothScroller);
    } else {
      callback && callback();
    }
  }

  window.requestAnimationFrame(smoothScroller);
}

function isArrayWithLength(test) {
  return Array.isArray(test) && !!test.length;
}
function createClassName(classNames) {
  return classNames.filter(x => x).join(' ');
}
function isPlainObject(test) {
  return test && typeof test === 'object' && !Array.isArray(test);
}
function createID(idPrefix, idSuffix) {
  if (!idPrefix) {
    return undefined;
  }

  if (!idSuffix) {
    return idPrefix;
  }

  return `${idPrefix}-${idSuffix}`;
}
const isTouchDevice = () => window.matchMedia('(pointer: coarse)').matches;
function trimAndFormatFilterStr(value, filterIsCaseSensitive) {
  const formatVal = value.replace(/^\s+|\s+$/g, '');
  return !filterIsCaseSensitive ? formatVal.toLowerCase() : formatVal;
}
function mergeDeep(target, source) {
  const output = Object.assign({}, target);
  Object.keys(source).forEach(key => {
    if (isPlainObject(source[key])) {
      output[key] = !(key in target) ? source[key] : mergeDeep(target[key], source[key]);
    } else {
      output[key] = source[key];
    }
  });
  return output;
}
function scrollMenuIntoViewOnOpen(menuEl, scrollMenuIntoView, handleOnMenuOpen) {
  if (!scrollMenuIntoView || !menuEl || !menuEl.getBoundingClientRect) {
    handleOnMenuOpen();
    return;
  }

  const {
    top: menuTop,
    bottom: menuBottom,
    height: menuHeight
  } = menuEl.getBoundingClientRect();
  const viewHeight = window.innerHeight;
  const viewSpaceBelow = viewHeight - menuTop;

  if (viewSpaceBelow >= menuHeight) {
    handleOnMenuOpen();
    return;
  }

  const scrollParent = getScrollParent(menuEl);
  const scrollTop = getScrollTop(scrollParent);
  const scrollSpaceBelow = scrollParent.getBoundingClientRect().height - scrollTop - menuTop;

  if (scrollSpaceBelow < menuHeight) {
    handleOnMenuOpen(scrollSpaceBelow);
    return;
  }

  const marginBottom = parseInt(getComputedStyle(menuEl).marginBottom || '0', 10);
  const scrollDown = menuBottom - viewHeight + scrollTop + marginBottom;
  smoothScrollTo(scrollParent, scrollDown, 300, handleOnMenuOpen);
}
function validateSetValueOption(option, menuOptions, getOptionValueCB) {
  if (option === null || option === undefined || Array.isArray(option)) {
    return undefined;
  }

  const optionValue = option && option !== Object(option) ? option : getOptionValueCB(option);
  return menuOptions.find(mOption => mOption.value === optionValue) || undefined;
}
const renderControlEmphasis = (boxShadow, borderColor, invalidColor, invalidFocus, boxShadowColor, focusedBorderColor, isFocused, isInvalid) => {
  if (isFocused) {
    return css`
      border-color: ${isInvalid ? invalidColor : focusedBorderColor};
      box-shadow: ${boxShadow} ${isInvalid ? invalidFocus : boxShadowColor};
    `;
  }

  return css`
    border-color: ${isInvalid ? invalidColor : borderColor};
  `;
};

const useMenuHeight = (menuRef, menuOpen, menuHeightDefault, scrollMenuIntoView, onMenuOpen, onMenuClose) => {
  const [menuHeight, setMenuHeight] = useState(menuHeightDefault);
  const handleOnMenuOpen = useCallback(availableSpace => {
    availableSpace && setMenuHeight(availableSpace);
    onMenuOpen && onMenuOpen();
  }, [onMenuOpen]);
  useEffect(() => {
    if (menuOpen) {
      scrollMenuIntoViewOnOpen(menuRef.current, scrollMenuIntoView, handleOnMenuOpen);
    } else {
      setMenuHeight(menuHeightDefault);
      onMenuClose && onMenuClose();
    }
  }, [menuRef, menuOpen, onMenuClose, handleOnMenuOpen, menuHeightDefault, scrollMenuIntoView]);
  return menuHeight;
};

const MENU_ITEM_SIZE_DEFAULT = 35;
const MENU_MAX_HEIGHT_DEFAULT = 300;
const NO_OPTIONS_MSG_DEFAULT = 'No options';
const PLACEHOLDER_DEFAULT = 'Select option..';
const OPTIONS_DEFAULT = [];
const NO_SELECTED_OPTION = Object.freeze({});
const FOCUSED_OPTION_DEFAULT = Object.freeze({
  index: -1
});
const SELECTED_OPTION_DEFAULT = Object.freeze({
  initFlag: true
});

const useMenuOptions = (options, debouncedInputValue, getOptionValueCB, getOptionLabelCB, getIsOptionDisabledCB, getFilterOptionStringCB, filterIsCaseSensitive) => {
  const [menuOptions, setMenuOptions] = useState(OPTIONS_DEFAULT);
  useEffect(() => {
    const createMenuOptions = () => {
      const cleanSearchInputValue = trimAndFormatFilterStr(debouncedInputValue, filterIsCaseSensitive);

      const optionSatisfiesFilter = menuOption => {
        const cleanStringifiedOption = trimAndFormatFilterStr(getFilterOptionStringCB(menuOption), filterIsCaseSensitive);
        return cleanStringifiedOption.indexOf(cleanSearchInputValue) > -1;
      };

      const parseMenuOption = data => {
        const menuOption = {
          data,
          label: getOptionLabelCB(data),
          value: getOptionValueCB(data)
        };

        if (cleanSearchInputValue && !optionSatisfiesFilter(menuOption)) {
          return undefined;
        }

        return Object.assign({}, menuOption, {}, getIsOptionDisabledCB(data) && {
          isDisabled: true
        });
      };

      return options.reduce((accumulator, data) => {
        const option = parseMenuOption(data);
        option && accumulator.push(option);
        return accumulator;
      }, []);
    };

    setMenuOptions(createMenuOptions() || OPTIONS_DEFAULT);
  }, [options, filterIsCaseSensitive, debouncedInputValue, getFilterOptionStringCB, getOptionValueCB, getOptionLabelCB, getIsOptionDisabledCB]);
  return menuOptions;
};

const attrPrefix = 'rfs';
const isNotProduction = process.env.NODE_ENV !== 'production';
const OPTION_CLS = `${attrPrefix}-option`;
const OPTION_FOCUSED_CLS = `${OPTION_CLS}-focused`;
const OPTION_SELECTED_CLS = `${OPTION_CLS}-selected`;
const OPTION_DISABLED_CLS = `${OPTION_CLS}-disabled`;
const CARET_ICON_CLS = `${attrPrefix}-caret-icon`;
const CLEAR_ICON_CLS = `${attrPrefix}-clear-icon`;
const AUTOSIZE_INPUT_CLS = `${attrPrefix}-autosize-input`;
const MENU_CONTAINER_CLS = `${attrPrefix}-menu-container`;
const SELECT_CONTAINER_CLS = `${attrPrefix}-select-container`;
const CONTROL_CONTAINER_CLS = `${attrPrefix}-control-container`;
const CLEAR_ICON_TESTID = isNotProduction ? CLEAR_ICON_CLS : undefined;
const AUTOSIZE_INPUT_TESTID = isNotProduction ? AUTOSIZE_INPUT_CLS : undefined;
const MENU_CONTAINER_TESTID = isNotProduction ? MENU_CONTAINER_CLS : undefined;
const SELECT_CONTAINER_TESTID = isNotProduction ? SELECT_CONTAINER_CLS : undefined;
const CONTROL_CONTAINER_TESTID = isNotProduction ? CONTROL_CONTAINER_CLS : undefined;

const Option = React.memo(({
  index,
  style,
  data: {
    idSuffix,
    menuOptions,
    selectOption,
    renderOptionLabel,
    focusedOptionIndex,
    selectedOptionValue
  }
}) => {
  const {
    data,
    value,
    label,
    isDisabled
  } = menuOptions[index];
  const isSelected = value === selectedOptionValue;
  const className = createClassName([OPTION_CLS, isDisabled && OPTION_DISABLED_CLS, isSelected && OPTION_SELECTED_CLS, index === focusedOptionIndex && OPTION_FOCUSED_CLS]);
  return React.createElement("div", {
    role: 'button',
    style: style,
    className: className,
    id: idSuffix ? createID(`${OPTION_CLS}-${index}`, idSuffix) : undefined,
    onClick: !isDisabled ? () => selectOption({
      data,
      value,
      label
    }, isSelected) : undefined
  }, renderOptionLabel(data));
}, areEqual);
Option.displayName = 'Option';

const NoOptionsMsg = styled.div`
  text-align: center;
  color: ${({
  theme
}) => theme.noOptions.color};
  margin: ${({
  theme
}) => theme.noOptions.margin};
  padding: ${({
  theme
}) => theme.noOptions.padding};
  font-size: ${({
  theme
}) => theme.noOptions.fontSize};
`;
const Menu = React.forwardRef(({
  width,
  itemSize,
  idSuffix,
  maxHeight,
  menuOptions,
  selectOption,
  noOptionsMsg,
  overscanCount,
  renderOptionLabel,
  focusedOptionIndex,
  selectedOptionValue
}, ref) => {
  const itemData = useMemo(() => ({
    idSuffix,
    menuOptions,
    selectOption,
    renderOptionLabel,
    focusedOptionIndex,
    selectedOptionValue
  }), [idSuffix, menuOptions, focusedOptionIndex, selectedOptionValue, selectOption, renderOptionLabel]);
  return React.createElement(Fragment, null, React.createElement(FixedSizeList, {
    ref: ref,
    width: width,
    itemSize: itemSize,
    itemData: itemData,
    overscanCount: overscanCount,
    itemCount: menuOptions.length,
    height: Math.min(maxHeight, menuOptions.length * itemSize)
  }, Option), !isArrayWithLength(menuOptions) && React.createElement(NoOptionsMsg, null, noOptionsMsg));
});
Menu.displayName = 'Menu';

const _valueDivCSS = css`
  top: 50%;
  overflow: hidden;
  position: absolute;
  white-space: nowrap;
  box-sizing: border-box;
  text-overflow: ellipsis;
  transform: translateY(-50%);
`;

const StyledValue = styled.div`
  ${_valueDivCSS}
  max-width: calc(100% - 0.5rem);
`;
const Placeholder = styled.div`
  ${_valueDivCSS}
  color: ${({
  theme
}) => theme.color.placeholder};
`;
const Value = React.memo(({
  inputValue,
  placeholder,
  renderOptionLabel,
  selectedOptionData
}) => {
  if (inputValue) {
    return null;
  }

  if (!selectedOptionData) {
    return React.createElement(Placeholder, null, placeholder);
  }

  return React.createElement(StyledValue, null, renderOptionLabel(selectedOptionData));
});
Value.displayName = 'Value';

const _inputMinWidthPx = 2;

const _wrapperDivStyle = Object.freeze({
  display: 'inline-block'
});

const _inputStaticAttributes = Object.freeze({
  type: 'text',
  spellCheck: false,
  autoCorrect: 'off',
  autoComplete: 'off',
  autoCapitalize: 'none',
  'aria-autocomplete': 'list',
  'data-testid': AUTOSIZE_INPUT_TESTID
});

const StyledSizer = styled.div`
  top: 0;
  left: 0;
  height: 0;
  overflow: scroll;
  white-space: pre;
  position: absolute;
  visibility: hidden;
  font-size: inherit;
  font-weight: inherit;
  font-family: inherit;
`;
const StyledAutosizeInput = styled.input`
  border: 0;
  outline: 0;
  padding: 0;
  cursor: text;
  background: 0;
  color: inherit;
  font-size: inherit;
  font-weight: inherit;
  font-family: inherit;
  box-sizing: content-box;

  :read-only {
    opacity: 0;
    cursor: default;
  }

  ${mediaQueryIsIE} {
    ::-ms-clear { 
      display: none; 
    }
  }
`;
const AutosizeInput = React.memo(React.forwardRef(({
  id,
  onBlur,
  onFocus,
  onChange,
  disabled,
  ariaLabel,
  inputValue,
  isSearchable,
  addClassNames,
  ariaLabelledBy
}, ref) => {
  const sizerRef = useRef(null);
  const [inputWidth, setInputWidth] = useState(_inputMinWidthPx);
  useEffect(() => {
    if (sizerRef.current) {
      setInputWidth(sizerRef.current.scrollWidth + _inputMinWidthPx);
    }
  }, [inputValue]);
  const inputAttributes = Object.assign({
    id,
    disabled
  }, _inputStaticAttributes, {
    'aria-label': ariaLabel,
    'aria-labelledby': ariaLabelledBy
  });
  return React.createElement("div", {
    style: _wrapperDivStyle
  }, React.createElement(StyledAutosizeInput, Object.assign({
    ref: ref,
    onBlur: onBlur,
    onFocus: onFocus,
    value: inputValue
  }, inputAttributes, {
    readOnly: !isSearchable,
    style: {
      width: inputWidth
    },
    onChange: isSearchable ? onChange : undefined,
    className: addClassNames ? AUTOSIZE_INPUT_CLS : undefined
  })), React.createElement(StyledSizer, {
    ref: sizerRef
  }, inputValue));
}));
AutosizeInput.displayName = 'AutosizeInput';

const bounce = keyframes`
  0%, 80%, 100% {
    transform: scale(0);
  } 40% {
    transform: scale(1.0);
  }
`;
const StyledLoadingDots = styled.div`
  display: flex;
  align-self: center;
  text-align: center;
  margin-right: 0.25rem;
  padding: ${({
  theme
}) => theme.loader.padding};

  > div {
    border-radius: 100%;
    display: inline-block;
    width: ${({
  theme
}) => theme.loader.size};
    height: ${({
  theme
}) => theme.loader.size};
    opacity: ${({
  theme
}) => theme.loader.opacity};
    background-color: ${({
  theme
}) => theme.loader.color};
    animation: ${bounce} 1.19s ease-in-out infinite;

    :nth-of-type(1) {
      animation-delay: -0.272s;
    }

    :nth-of-type(2) {
      animation-delay: -0.136s;
    }
  }
`;

const LoadingDots = () => React.createElement(StyledLoadingDots, {
  "aria-hidden": 'true'
}, React.createElement("div", null), React.createElement("div", null), React.createElement("div", null));

const IndicatorIconsWrapper = styled.div`
  display: flex;
  flex-shrink: 0;
  align-items: center;
  align-self: stretch;
  box-sizing: border-box;
`;
const IndicatorIcon = styled.div`
  display: flex;
  box-sizing: border-box;
  color: ${({
  theme
}) => theme.icon.color};
  padding: ${({
  theme
}) => theme.icon.padding};

  :hover {
    color: ${({
  theme
}) => theme.icon.hoverColor};
  }
`;
const Clear = styled.div`
  overflow: hidden;
  ${fadeInAnimationCss}
  font-size: ${({
  theme
}) => theme.icon.clear.fontSize};
  transition: ${({
  theme
}) => theme.icon.clear.transition};
  font-weight: ${({
  theme
}) => theme.icon.clear.fontWeight};
  font-family: ${({
  theme
}) => theme.icon.clear.fontFamily};
`;
const Caret = styled.div`
  transition: ${({
  theme
}) => theme.icon.caret.transition};
  border-top: ${({
  theme
}) => theme.icon.caret.size} dashed;
  border-left: ${({
  theme
}) => theme.icon.caret.size} solid transparent;
  border-right: ${({
  theme
}) => theme.icon.caret.size} solid transparent;

  ${({
  menuOpen,
  isInvalid,
  theme: {
    color
  }
}) => menuOpen && css`
      transform: rotate(180deg);
      color: ${isInvalid ? color.invalid : color.caretActive || color.accent};
    `}
`;
const Separator = styled.div`
  width: 1px;
  margin: 0.5rem 0;
  align-self: stretch;
  box-sizing: border-box;
  background-color: ${({
  theme: {
    color
  }
}) => color.iconSeparator || color.border};
`;
const IndicatorIcons = React.memo(({
  menuOpen,
  clearIcon,
  caretIcon,
  isInvalid,
  showClear,
  isLoading,
  addClassNames,
  onClearMouseDown
}) => React.createElement(IndicatorIconsWrapper, null, showClear && !isLoading && React.createElement(IndicatorIcon, {
  "aria-hidden": 'true',
  onTouchEnd: onClearMouseDown,
  onMouseDown: onClearMouseDown,
  "data-testid": CLEAR_ICON_TESTID
}, clearIcon || React.createElement(Clear, {
  className: addClassNames ? CLEAR_ICON_CLS : undefined
}, "X")), isLoading && React.createElement(LoadingDots, null), React.createElement(Separator, null), React.createElement(IndicatorIcon, {
  "aria-hidden": 'true'
}, caretIcon || React.createElement(Caret, {
  menuOpen: menuOpen,
  isInvalid: isInvalid,
  className: addClassNames ? CARET_ICON_CLS : undefined
}))));
IndicatorIcons.displayName = 'IndicatorIcons';

const A11yText = styled.span`
  border: 0;
  padding: 0;
  width: 1px;
  height: 1px;
  margin: -1px;
  overflow: hidden;
  position: absolute;
  white-space: nowrap;
  clip: rect(0 0 0 0);
  clip-path: inset(50%);
`;

const AriaLiveRegion = ({
  menuOpen,
  ariaLabel,
  inputValue,
  optionCount,
  isSearchable,
  focusedOptionLabel,
  focusedOptionIndex,
  selectedOptionLabel,
  isFocusedOptionDisabled
}) => {
  const selectedOptionMsg = `Selected option: ${selectedOptionLabel || 'N/A'}`;
  const optionsMsg = `${optionCount} result(s) available${inputValue ? ` for search input ${inputValue}` : ''}.`;
  const focusedMsg = `Focused option: ${focusedOptionLabel || 'N/A'}${isFocusedOptionDisabled ? ' - disabled' : ''}, ${focusedOptionIndex + 1} of ${optionCount}.`;
  const menuMsg = menuOpen ? 'Use Up and Down arrow keys to choose options, press Enter or Tab to select the currently focused option, press Escape to close the menu.' : `${ariaLabel || 'Select'} is focused${isSearchable ? ', type to filter options' : ''}, press Down arrow key to open the menu.`;
  return React.createElement(A11yText, {
    "aria-live": 'polite'
  }, React.createElement("p", null, selectedOptionMsg), React.createElement("p", null, focusedMsg + optionsMsg + menuMsg));
};

const _indexPositionEnum = Object.freeze({
  UP: 0,
  DOWN: 1,
  FIRST: 2,
  LAST: 3
});

const SelectWrapper = styled.div`
  position: relative;
  box-sizing: border-box;
  ${({
  theme
}) => theme.color.textColor && css`color: ${theme.color.textColor};`}
  ${({
  theme
}) => theme.select.fontSize && css`font-size: ${theme.select.fontSize};`}
`;
const ValueWrapper = styled.div`
  flex: 1 1 0%;
  display: flex;
  flex-wrap: wrap;
  overflow: hidden;
  position: relative;
  align-items: center;
  box-sizing: border-box;
  padding: ${({
  theme
}) => theme.control.padding};
`;
const ControlWrapper = styled.div`
  outline: 0;
  display: flex;
  flex-wrap: wrap;
  cursor: default;
  position: relative;
  align-items: center;
  box-sizing: border-box;
  justify-content: space-between;

  ${({
  theme: {
    control
  }
}) => css`
    height: ${control.height};
    transition: ${control.transition};
    border-style: ${control.borderStyle};
    border-width: ${control.borderWidth};
    border-radius: ${control.borderRadius};
    ${control.backgroundColor && `background-color: ${control.backgroundColor};`}
  `}

  ${({
  isDisabled,
  theme
}) => isDisabled && css`
      pointer-events: none;
      background-color: ${theme.color.disabled};
    `}

  ${({
  isFocused,
  isInvalid,
  theme
}) => renderControlEmphasis(theme.control.boxShadow, theme.color.border, theme.color.invalid, theme.color.invalidFocus, theme.control.boxShadowColor, theme.control.focusedBorderColor, isFocused, isInvalid)}
`;
const MenuWrapper = styled.div`
  z-index: 999;
  position: absolute;
  ${fadeInAnimationCss}
  
  ${({
  hideMenu,
  theme: {
    menu
  }
}) => css`
    width: ${menu.width};
    margin: ${menu.margin};
    padding: ${menu.padding};
    box-shadow: ${menu.boxShadow};
    border-radius: ${menu.borderRadius};
    background-color: ${menu.backgroundColor};
    ${hideMenu && 'display: none;'}
  `}

  .${OPTION_CLS} {
    display: block;
    overflow: hidden;
    user-select: none;
    white-space: nowrap;
    text-overflow: ellipsis;
    -webkit-tap-highlight-color: transparent;
    padding: ${({
  theme
}) => theme.menu.option.padding};
    text-align: ${({
  theme
}) => theme.menu.option.textAlign};

    &.${OPTION_FOCUSED_CLS},
    &:hover:not(.${OPTION_DISABLED_CLS}):not(.${OPTION_SELECTED_CLS}) {
      background-color: ${({
  theme
}) => theme.menu.option.focusedBgColor};
    }

    &.${OPTION_SELECTED_CLS} {
      color: ${({
  theme
}) => theme.menu.option.selectedColor};
      background-color: ${({
  theme
}) => theme.menu.option.selectedBgColor};
    }

    &.${OPTION_DISABLED_CLS} {
      opacity: 0.35;
    }
  }
`;
const Select = React.forwardRef(({
  inputId,
  selectId,
  idSuffix,
  isLoading,
  onKeyDown,
  clearIcon,
  caretIcon,
  isInvalid,
  ariaLabel,
  menuWidth,
  isDisabled,
  inputDelay,
  onMenuOpen,
  onMenuClose,
  onInputBlur,
  isClearable,
  onInputFocus,
  initialValue,
  addClassNames,
  ariaLabelledBy,
  onOptionChange,
  getOptionLabel,
  getOptionValue,
  openMenuOnFocus,
  isAriaLiveEnabled,
  menuOverscanCount,
  blurInputOnSelect,
  renderOptionLabel,
  getIsOptionDisabled,
  filterIsCaseSensitive,
  getFilterOptionString,
  isSearchable = true,
  tabSelectsOption = true,
  closeMenuOnSelect = true,
  scrollMenuIntoView = true,
  backspaceClearsValue = true,
  options = OPTIONS_DEFAULT,
  placeholder = PLACEHOLDER_DEFAULT,
  noOptionsMsg = NO_OPTIONS_MSG_DEFAULT,
  menuItemSize = MENU_ITEM_SIZE_DEFAULT,
  menuMaxHeight = MENU_MAX_HEIGHT_DEFAULT
}, ref) => {
  const prevMenuOptionsLen = useRef();
  const listRef = useRef(null);
  const menuRef = useRef(null);
  const inputRef = useRef(null);
  const [inputValue, setInputValue] = useState('');
  const [menuOpen, setMenuOpen] = useState(false);
  const [isFocused, setIsFocused] = useState(false);
  const [focusedOption, setFocusedOption] = useState(FOCUSED_OPTION_DEFAULT);
  const [selectedOption, setSelectedOption] = useState(SELECTED_OPTION_DEFAULT);
  const debouncedInputValue = useDebounce(inputValue, inputDelay);
  const {
    data: selectedOptionData,
    value: selectedOptionValue,
    label: selectedOptionLabel,
    initFlag: canAttemptInitValSet
  } = selectedOption;
  const {
    data: focusedOptionData,
    value: focusedOptionValue,
    label: focusedOptionLabel,
    index: focusedOptionIndex,
    isDisabled: isFocusedOptionDisabled
  } = focusedOption;
  const getOptionLabelCB = useCallback(data => {
    return getOptionLabel ? getOptionLabel(data) : data.label;
  }, [getOptionLabel]);
  const getOptionValueCB = useCallback(data => {
    return getOptionValue ? getOptionValue(data) : data.value;
  }, [getOptionValue]);
  const getIsOptionDisabledCB = useCallback(data => {
    return getIsOptionDisabled ? getIsOptionDisabled(data) : !!data.isDisabled;
  }, [getIsOptionDisabled]);
  const renderOptionLabelCB = useCallback(data => {
    return renderOptionLabel ? renderOptionLabel(data) : getOptionLabelCB(data);
  }, [renderOptionLabel, getOptionLabelCB]);
  const getFilterOptionStringCB = useCallback(option => {
    return getFilterOptionString ? getFilterOptionString(option) : String(option.label);
  }, [getFilterOptionString]);
  const menuHeight = useMenuHeight(menuRef, menuOpen, menuMaxHeight, scrollMenuIntoView, onMenuOpen, onMenuClose);
  const menuOptions = useMenuOptions(options, debouncedInputValue, getOptionValueCB, getOptionLabelCB, getIsOptionDisabledCB, getFilterOptionStringCB, filterIsCaseSensitive);

  const blurInput = () => {
    inputRef.current && inputRef.current.blur();
  };

  const focusInput = () => {
    inputRef.current && inputRef.current.focus();
  };

  const scrollToItemIndex = index => {
    listRef.current && listRef.current.scrollToItem(index);
  };

  const selectOption = useCallback((option, isSelected) => {
    if (isSelected) {
      setMenuOpen(false);
    } else {
      setSelectedOption(option || NO_SELECTED_OPTION);
    }
  }, []);
  const openMenuAndFocusOption = useCallback(position => {
    if (!isArrayWithLength(menuOptions)) {
      setMenuOpen(true);
      return;
    }

    const selectedIndex = selectedOption.value ? menuOptions.findIndex(option => option.value === selectedOption.value) : -1;
    const index = selectedIndex > -1 ? selectedIndex : position === _indexPositionEnum.FIRST ? 0 : menuOptions.length - 1;
    setMenuOpen(true);
    setFocusedOption(Object.assign({
      index
    }, menuOptions[index]));
    scrollToItemIndex(index);
  }, [menuOptions, selectedOption]);
  useImperativeHandle(ref, () => ({
    blur: blurInput,
    focus: focusInput,
    clearValue: () => {
      selectOption();
      setFocusedOption(FOCUSED_OPTION_DEFAULT);
    },
    setValue: option => {
      const validOptionOrUndefined = validateSetValueOption(option, menuOptions, getOptionValueCB);
      selectOption(validOptionOrUndefined);
    }
  }));
  useEffect(() => {
    if (isFocused && openMenuOnFocus) {
      openMenuAndFocusOption(_indexPositionEnum.FIRST);
    }
  }, [isFocused, openMenuOnFocus, openMenuAndFocusOption]);
  useEffect(() => {
    if (canAttemptInitValSet && isPlainObject(initialValue)) {
      selectOption({
        data: initialValue,
        value: getOptionValueCB(initialValue),
        label: getOptionLabelCB(initialValue)
      });
    }
  }, [selectOption, initialValue, canAttemptInitValSet, getOptionValueCB, getOptionLabelCB]);
  useEffect(() => {
    onOptionChange && onOptionChange(selectedOptionData || null);
    !selectedOptionData && setFocusedOption(FOCUSED_OPTION_DEFAULT);

    if (blurInputOnSelect && selectedOptionData) {
      blurInput();
    } else if (closeMenuOnSelect) {
      setMenuOpen(false);
      setInputValue('');
    }
  }, [selectedOptionData, onOptionChange, closeMenuOnSelect, blurInputOnSelect]);
  useEffect(() => {
    if (menuOptions.length === 1 || !!menuOptions.length && (menuOptions.length !== options.length || prevMenuOptionsLen.current === 0)) {
      setFocusedOption(Object.assign({
        index: 0
      }, menuOptions[0]));
      scrollToItemIndex(0);
    }

    prevMenuOptionsLen.current = menuOptions.length;
  }, [options, menuOptions]);

  const selectOptionFromFocused = () => {
    if (!focusedOptionData || isFocusedOptionDisabled) {
      return;
    }

    if (closeMenuOnSelect && selectedOptionValue === focusedOptionValue) {
      setMenuOpen(false);
    } else {
      selectOption({
        data: focusedOptionData,
        value: focusedOptionValue,
        label: focusedOptionLabel
      });
    }
  };

  const focusOptionOnArrowKey = position => {
    if (!isArrayWithLength(menuOptions)) {
      return;
    }

    const index = position === _indexPositionEnum.DOWN ? (focusedOptionIndex + 1) % menuOptions.length : focusedOptionIndex > 0 ? focusedOptionIndex - 1 : menuOptions.length - 1;
    setFocusedOption(Object.assign({
      index
    }, menuOptions[index]));
    scrollToItemIndex(index);
  };

  const handleOnKeyDown = e => {
    if (isDisabled) {
      return;
    }

    if (onKeyDown) {
      onKeyDown(e);

      if (e.defaultPrevented) {
        return;
      }
    }

    switch (e.key) {
      case 'ArrowDown':
        if (menuOpen) {
          focusOptionOnArrowKey(_indexPositionEnum.DOWN);
        } else {
          openMenuAndFocusOption(_indexPositionEnum.FIRST);
        }

        break;

      case 'ArrowUp':
        if (menuOpen) {
          focusOptionOnArrowKey(_indexPositionEnum.UP);
        } else {
          openMenuAndFocusOption(_indexPositionEnum.LAST);
        }

        break;

      case ' ':
        if (inputValue) {
          return;
        } else if (!menuOpen) {
          openMenuAndFocusOption(_indexPositionEnum.FIRST);
        } else {
          selectOptionFromFocused();
        }

        break;

      case 'Enter':
        if (e.keyCode !== 229) {
          selectOptionFromFocused();
        }

        break;

      case 'Escape':
        setMenuOpen(false);
        setInputValue('');
        break;

      case 'Tab':
        if (e.shiftKey || !menuOpen || !focusedOptionData) {
          return;
        } else if (tabSelectsOption) {
          selectOptionFromFocused();
        } else {
          setMenuOpen(false);
          setInputValue('');
        }

        break;

      case 'Delete':
      case 'Backspace':
        if (inputValue || !backspaceClearsValue || !isClearable) {
          return;
        }

        selectOption();
        break;

      default:
        return;
    }

    e.preventDefault();
  };

  const handleOnControlMouseDown = e => {
    if (isDisabled) {
      return;
    }

    if (!isFocused) {
      focusInput();
    }

    if (!menuOpen) {
      openMenuAndFocusOption(_indexPositionEnum.FIRST);
    } else if (e.currentTarget.tagName !== 'INPUT') {
      setMenuOpen(false);
      inputValue && setInputValue('');
    }

    if (e.currentTarget.tagName !== 'INPUT') {
      e.preventDefault();
    }
  };

  const handleOnMenuMouseDown = e => {
    e.stopPropagation();
    e.preventDefault();
    focusInput();
  };

  const handleOnInputBlur = useCallback(e => {
    onInputBlur && onInputBlur(e);
    setIsFocused(false);
    setMenuOpen(false);
    setInputValue('');
  }, [onInputBlur]);
  const handleOnInputFocus = useCallback(e => {
    onInputFocus && onInputFocus(e);
    setIsFocused(true);
  }, [onInputFocus]);
  const handleOnInputChange = useCallback(e => {
    setMenuOpen(true);
    setInputValue(e.currentTarget.value || '');
  }, []);
  const handleOnClearMouseDown = useCallback(e => {
    e.stopPropagation();
    e.type === 'mousedown' && e.preventDefault();
    setSelectedOption(NO_SELECTED_OPTION);
    focusInput();
  }, []);
  return React.createElement(SelectWrapper, {
    id: selectId,
    onKeyDown: handleOnKeyDown,
    "data-testid": SELECT_CONTAINER_TESTID,
    className: addClassNames ? SELECT_CONTAINER_CLS : undefined
  }, React.createElement(ControlWrapper, {
    isInvalid: isInvalid,
    isFocused: isFocused,
    isDisabled: isDisabled,
    onTouchEnd: handleOnControlMouseDown,
    onMouseDown: handleOnControlMouseDown,
    "data-testid": CONTROL_CONTAINER_TESTID,
    className: addClassNames ? CONTROL_CONTAINER_CLS : undefined
  }, React.createElement(ValueWrapper, null, React.createElement(Value, {
    inputValue: inputValue,
    placeholder: placeholder,
    selectedOptionData: selectedOptionData,
    renderOptionLabel: renderOptionLabelCB
  }), React.createElement(AutosizeInput, {
    id: inputId,
    ref: inputRef,
    disabled: isDisabled,
    ariaLabel: ariaLabel,
    inputValue: inputValue,
    onBlur: handleOnInputBlur,
    isSearchable: isSearchable,
    onFocus: handleOnInputFocus,
    addClassNames: addClassNames,
    onChange: handleOnInputChange,
    ariaLabelledBy: ariaLabelledBy
  })), React.createElement(IndicatorIcons, {
    menuOpen: menuOpen,
    clearIcon: clearIcon,
    caretIcon: caretIcon,
    isInvalid: isInvalid,
    isLoading: isLoading,
    addClassNames: addClassNames,
    onClearMouseDown: handleOnClearMouseDown,
    showClear: !!(isClearable && !isDisabled && selectedOptionData)
  })), React.createElement(MenuWrapper, {
    ref: menuRef,
    hideMenu: !menuOpen,
    onMouseDown: handleOnMenuMouseDown,
    "data-testid": MENU_CONTAINER_TESTID,
    className: addClassNames ? MENU_CONTAINER_CLS : undefined
  }, React.createElement(Menu, {
    ref: listRef,
    width: menuWidth,
    idSuffix: idSuffix,
    maxHeight: menuHeight,
    itemSize: menuItemSize,
    menuOptions: menuOptions,
    noOptionsMsg: noOptionsMsg,
    selectOption: selectOption,
    overscanCount: menuOverscanCount,
    focusedOptionIndex: focusedOptionIndex,
    renderOptionLabel: renderOptionLabelCB,
    selectedOptionValue: selectedOptionValue
  })), isAriaLiveEnabled && isFocused && React.createElement(AriaLiveRegion, {
    menuOpen: menuOpen,
    ariaLabel: ariaLabel,
    inputValue: inputValue,
    isSearchable: isSearchable,
    optionCount: menuOptions.length,
    focusedOptionLabel: focusedOptionLabel,
    focusedOptionIndex: focusedOptionIndex,
    selectedOptionLabel: selectedOptionLabel,
    isFocusedOptionDisabled: isFocusedOptionDisabled
  }));
});
Select.displayName = 'Select';

var DefaultThemeObj = Object.freeze({
  color: {
    accent: '#007bff',
    border: '#ced4da',
    invalid: '#dc3545',
    disabled: '#e9ecef',
    placeholder: '#6E7276',
    invalidFocus: 'rgba(220, 53, 69, 0.25)'
  },
  select: {},
  loader: {
    opacity: '0.42',
    size: '0.625rem',
    color: '#007bff',
    padding: '0.375rem 0.75rem'
  },
  icon: {
    color: '#cccccc',
    padding: '0.5rem 0.9375rem',
    hoverColor: '#A6A6A6',
    clear: {
      fontWeight: 900,
      fontSize: '0.85em',
      transition: 'color 0.15s ease-in-out',
      fontFamily: '"Helvetica", "Arial", sans-serif'
    },
    caret: {
      size: '7px',
      transition: 'transform 0.225s ease-in-out, color 0.15s ease-in-out'
    }
  },
  control: {
    borderWidth: '1px',
    borderStyle: 'solid',
    borderRadius: '0.25rem',
    boxShadow: '0 0 0 0.2rem',
    padding: '0.375rem 0.75rem',
    height: 'calc(1.5em + 0.75rem + 2px)',
    boxShadowColor: 'rgba(0, 123, 255, 0.25)',
    focusedBorderColor: 'rgba(0, 123, 255, 0.75)',
    transition: 'box-shadow 0.15s ease-in-out, border-color 0.15s ease-in-out'
  },
  menu: {
    width: '100%',
    margin: '0.5rem 0',
    padding: '0.15rem 0',
    borderRadius: '0.25rem',
    backgroundColor: '#fff',
    boxShadow: '0 0 0 1px rgba(0, 0, 0, 0.1), 0 4px 11px rgba(0, 0, 0, 0.1)',
    option: {
      textAlign: 'left',
      selectedColor: '#fff',
      selectedBgColor: '#007bff',
      padding: '0.375rem 0.75rem',
      focusedBgColor: 'rgba(0, 123, 255, 0.20)'
    }
  },
  noOptions: {
    fontSize: '1.25rem',
    margin: '0.25rem 0',
    color: 'hsl(0, 0%, 60%)',
    padding: '0.375rem 0.75rem'
  }
});

const SelectWrapper$1 = React.forwardRef((_ref, ref) => {
  let {
    inputId,
    selectId,
    idSuffix,
    menuWidth,
    themeConfig,
    blurInputOnSelect
  } = _ref,
      rest = _objectWithoutPropertiesLoose(_ref, ["inputId", "selectId", "idSuffix", "menuWidth", "themeConfig", "blurInputOnSelect"]);

  const theme = useMemo(() => {
    return !isPlainObject(themeConfig) ? Object.assign({}, DefaultThemeObj) : mergeDeep(DefaultThemeObj, themeConfig);
  }, [themeConfig]);
  const blurInputOnSelectOrDefault = typeof blurInputOnSelect === 'boolean' ? blurInputOnSelect : isTouchDevice();
  const selectProps = Object.assign({}, rest, {
    idSuffix,
    inputId: createID(inputId, idSuffix),
    selectId: createID(selectId, idSuffix),
    menuWidth: menuWidth || theme.menu.width,
    blurInputOnSelect: blurInputOnSelectOrDefault
  });
  return React.createElement(ThemeProvider, {
    theme: theme
  }, React.createElement(Select, Object.assign({
    ref: ref
  }, selectProps)));
});
SelectWrapper$1.displayName = 'SelectWrapper';

export { SelectWrapper$1 as Select };
